/* $Id: rtl818x.h,v 1.1 2006/06/30 18:48:54 bsd3 Exp $ */

/*- 
 * Driver for Realtek's RTL8180L/8185L 802.11(a)/b/g MAC/BBP chipset
 * based wireless cards, see README for more details.
 *
 * Copyright (C) Andrea Merello 2004-2005  <andreamrl@tiscali.it> 
 *   
 * Parts of this driver are based on the GPL portion of the Realtek's 
 * official driver
 *   
 * Parts of this driver are based on the rtl8180 driver skeleton from
 * Patric Schenke, Andres Salomon and others.
 *   
 * Parts of this driver are based on the Intel Pro Wireless 2100 GPL
 * driver.
 *   
 * Thanks to the authors of all above said projects and the Ndiswrapper
 * project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifndef RTL818XH
#define RTL818XH

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/netdevice.h>
#include <linux/pci.h>
#include <linux/etherdevice.h>
#include <linux/delay.h>
#include <linux/rtnetlink.h>	/* for rtnl_lock() */
#include <linux/wireless.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>	/* Necessary because we use the proc fs */
#include <linux/if_arp.h>

#include <net/ieee80211.h>

#include <asm/io.h>
#include <asm/semaphore.h>

#define DRV_AUTHOR "Andrea Merello <andreamrl@tiscali.it>"
#define DRV_DESCRIPTION "Driver for Realtek's RTL8180/RTL8185 802.11 Wireless (Mini-)PCI/CardBus ethernet cards"
#define DRV_NAME "rtl818x"
#ifndef DRV_VERSION
#define DRV_VERSION "1.0.0-b1"
#endif

/* The debug system described here is drived from ipw2200 driver, I think;
 * anyway I have taken it from ieee80211.h (Dr Balwinder S Dheeman) */
#ifdef CONFIG_RTL818X_DEBUG
extern u32 rtl818x_debug_level;

#define RTL818X_DEBUG(lvl, fmt, args...) \
do { if (rtl818x_debug_level & (lvl)) \
  printk(KERN_DEBUG "ieee80211: %c %s " fmt, \
         in_interrupt() ? 'I' : 'U', __FUNCTION__ , ## args); } while (0)
#else
#define RTL818X_DEBUG(lvl, fmt, args...) do {} while (0)
#endif	/* CONFIG_RTL818X_DEBUG */

/* To use the debug system;
 *
 * If you are defining a new debug classification, simply add it to the #define
 * list here in the form of:
 *
 * #define RTL818X_DL_xxxx VALUE
 *
 * shifting value to the left one bit from the previous entry.  xxxx should be
 * the name of the classification (for example, WEP)
 *
 * You then need to either add a IEEE80211_xxxx_DEBUG() macro definition for your
 * classification, or use RTL818X_DEBUG(RTL818X_DL_xxxx, ...) whenever you want
 * to send output to that classification.
 *
 * To add your debug level to the list of levels seen when you perform
 *
 * % cat /proc/net/rtl818x/debug_level
 *
 * you simply need to add your entry to the rtl818x_debug_levels array.
 *
 * If you do not see debug_level in /proc/net/rtl818x then you do not have
 * CONFIG_RTL818X_DEBUG defined in your kernel configuration */

#define RTL818X_DL_INFO		(1<< 0)
#define RTL818X_DL_EPROM	(1<< 1)
#define RTL818X_DL_MGMT		(1<< 3)	/* Setup IRQ, RX, TX, WX and, or any
					   other */
#define RTL818X_DL_REGS		(1<< 4)
#define RTL818X_DL_ALLOC	(1<< 5)	/* Both RX and TX, if RX or TX bit is 
					   also set */
#define RTL818X_DL_FILL		(1<< 6)	/* ZERO_RX or TX_FILLDESC */
#define RTL818X_DL_FRAG		(1<< 7)	/* " */
#define RTL818X_DL_PHY		(1<< 8)
#define RTL818X_DL_RING		(1<< 9)	/* " */
#define RTL818X_DL_VERB		(1<<10)	/* SKB, VERBOSE, DESC and DESC2 */
#define RTL818X_DL_RX		(1<<11)
#define RTL818X_DL_TX		(1<<12)

#define RTL818X_DEBUG_INFO(f, a...) RTL818X_DEBUG(RTL818X_DL_INFO, f "\n", ## a)
#define RTL818X_DEBUG_EPROM(f, a...) RTL818X_DEBUG(RTL818X_DL_EPROM, f "\n", ## a)
#define RTL818X_DEBUG_MGMT(f, a...) RTL818X_DEBUG(RTL818X_DL_MGMT, f "\n", ## a)
#define RTL818X_DEBUG_REGS(f, a...) RTL818X_DEBUG(RTL818X_DL_RESG, f "\n", ## a)
#define RTL818X_DEBUG_ALLOC(f, a...) RTL818X_DEBUG(RTL818X_DL_ALLOC, f "\n", ## a)
#define RTL818X_DEBUG_FILL(f, a...) RTL818X_DEBUG(RTL818X_DL_FILL, f "\n", ## a)
#define RTL818X_DEBUG_FRAG(f, a...) RTL818X_DEBUG(RTL818X_DL_FRAG, f "\n", ## a)
#define RTL818X_DEBUG_PHY(f, a...) RTL818X_DEBUG(RTL818X_DL_PHY, f "\n", ## a)
#define RTL818X_DEBUG_RING(f, a...) RTL818X_DEBUG(RTL818X_DL_RING, f "\n", ## a)
#define RTL818X_DEBUG_VERB(f, a...) RTL818X_DEBUG(RTL818X_DL_VERB, f "\n", ## a)
#define RTL818X_DEBUG_RX(f, a...) RTL818X_DEBUG(RTL818X_DL_RX, f "\n", ## a)
#define RTL818X_DEBUG_TX(f, a...) RTL818X_DEBUG(RTL818X_DL_TX, f "\n", ## a)

/* useful macros */
#define RTL818X_INFO(f, a...) printk(KERN_INFO DRV_NAME ": " f "\n", ## a)
#define RTL818X_ERROR(f, a...) printk(KERN_ERR DRV_NAME ": (EE) " f "\n", ## a)
#define RTL818X_NOTICE(f, a...) printk(KERN_NOTICE DRV_NAME ": (!!) " f "\n", ## a)
#define RTL818X_WARN(f, a...) printk(KERN_WARNING DRV_NAME ": (WW) " f "\n", ## a)

#define EPROM_93c46 0
#define EPROM_93c56 1

#define RTL_IOCTL_WPA_SUPPLICANT SIOCIWFIRSTPRIV+30

#define DEFAULT_BEACON_ESSID "RTL818XL"
#define DEFAULT_BEACON_INTERVAL	0x64U
#define DEFAULT_FRAG_THRESHOLD MAX_FRAG_THRESHOLD
#define DEFAULT_RETRY_DATA 7
#define DEFAULT_RETRY_RTS 7
#define DEFAULT_SSID ""
#define PRISM_HDR_SIZE 64

typedef struct buffer {
	struct buffer *next;
	u32 *buf;
	dma_addr_t dma;
} buffer;

typedef struct Stats {
	unsigned long txrdu;
	unsigned long rxrdu;
	unsigned long rxnolast;
	unsigned long rxnodata;
	// unsigned long rxreset;
	// unsigned long rxwrkaround;
	unsigned long rxnopointer;
	unsigned long txnperr;
	unsigned long txresumed;
	unsigned long rxerr;
	unsigned long rxoverflow;
	unsigned long rxint;
	unsigned long txnpokint;
	unsigned long txhpokint;
	unsigned long txhperr;
	unsigned long ints;
	unsigned long shints;
	unsigned long txoverflow;
	unsigned long rxdmafail;
	unsigned long txbeacon;
	unsigned long txbeaconerr;
	unsigned long txlpokint;
	unsigned long txlperr;
} Stats;

typedef struct rtl818x_priv {
	struct pci_dev *pdev;

	short epromtype;
	int irq;
	struct ieee80211_device *ieee80211;

	short card_8185;	/* O: rtl8180, 1:rtl8185 V B/C, 2:rtl8185 V D 
				 */
	short card_8185_Bversion;	/* if TCR reports card V B/C this
					   discriminates */
	short phy_ver;		/* meaningful for rtl8225 1:A 2:B 3:C */
	short enable_gpio0;
	enum card_type { PCI, MINIPCI, CARDBUS,
		USB /* rtl8187 */
	} card_type;
	short hw_plcp_len;

	spinlock_t irq_lock;
	spinlock_t irq_th_lock;
	spinlock_t tx_lock;
	spinlock_t ps_lock;

	u16 irq_mask;
	short irq_enabled;
	struct net_device *dev;
	short chan;
	short sens;
	short max_sens;
	u8 chtxpwr[15];		/* channels from 1 to 14, 0 not used */
	u8 chtxpwr_ofdm[15];	/* channels from 1 to 14, 0 not used */
	// u8 challow[15]; /* channels from 1 to 14, 0 not used */
	short up;
	short crcmon;		/* if 1 allow bad crc frame reception in */
	/* monitor mode */
	short prism_hdr;

	struct timer_list scan_timer;
	// short scanpending;
	// short stopscan;
	spinlock_t scan_lock;
	u8 active_probe;
	// u8 active_scan_num;
	struct semaphore wx_sem;
	short hw_wep;

	short digphy;
	short antb;
	short diversity;
	u8 cs_treshold;
	short rcr_csense;
	short rf_chip;
	u32 key0[4];
	short (*rf_set_sens) (struct net_device * dev, short sens);
	void (*rf_set_chan) (struct net_device * dev, short ch);
	void (*rf_close) (struct net_device * dev);
	void (*rf_init) (struct net_device * dev);
	void (*rf_sleep) (struct net_device * dev);
	void (*rf_wakeup) (struct net_device * dev);
	// short rate;
	short promisc;

	/* stats */
	struct Stats stats;
	struct iw_statistics wstats;
	struct proc_dir_entry *dir_dev;

	/* RX stuff */
	u32 *rxring;
	u32 *rxringtail;
	dma_addr_t rxringdma;
	struct buffer *rxbuffer;
	struct buffer *rxbufferhead;
	int rxringcount;
	u16 rxbuffersize;

	struct sk_buff *rx_skb;
	short rx_skb_complete;
	u32 rx_prevlen;

	/* TX stuff */
	u32 *txlpring;
	u32 *txhpring;
	u32 *txnpring;
	dma_addr_t txlpringdma;
	dma_addr_t txhpringdma;
	dma_addr_t txnpringdma;
	u32 *txlpringtail;
	u32 *txhpringtail;
	u32 *txnpringtail;
	u32 *txlpringhead;
	u32 *txhpringhead;
	u32 *txnpringhead;
	struct buffer *txlpbufs;
	struct buffer *txhpbufs;
	struct buffer *txnpbufs;
	struct buffer *txlpbufstail;
	struct buffer *txhpbufstail;
	struct buffer *txnpbufstail;
	int txringcount;
	int txbuffsize;
	// struct tx_pendingbuf txnp_pending;
	// struct tasklet_struct irq_tx_tasklet;
	struct tasklet_struct irq_rx_tasklet;
	u8 dma_poll_mask;
	// short tx_suspend;

	/* adhoc/master mode stuff */
	u32 *txbeaconringtail;
	dma_addr_t txbeaconringdma;
	u32 *txbeaconring;
	int txbeaconcount;
	struct buffer *txbeaconbufs;
	struct buffer *txbeaconbufstail;
	// char *master_essid;
	// u16 master_beaconinterval;
	// u32 master_beaconsize;
	// u16 beacon_interval;

	u8 retry_data;
	u8 retry_rts;

	// short wq_hurryup;
	// struct workqueue_struct *workqueue;
	struct work_struct reset_wq;
	short ack_tx_to_ieee;
} rtl818x_priv;

#define BEACON_PRIORITY 3
#define HI_PRIORITY 2
#define LOW_PRIORITY 1
#define NORM_PRIORITY 0

short rtl818x_tx(struct net_device *dev, u32 * skbuf, int len,
		 int priority, short morefrag, short fragdesc, int rate);

u8 read_nic_byte(struct net_device *dev, int x);
u32 read_nic_dword(struct net_device *dev, int x);
u16 read_nic_word(struct net_device *dev, int x);
void write_nic_byte(struct net_device *dev, int x, u8 y);
void write_nic_word(struct net_device *dev, int x, u16 y);
void write_nic_dword(struct net_device *dev, int x, u32 y);
void force_pci_posting(struct net_device *dev);

void rtl818x_rtx_disable(struct net_device *);
void rtl818x_rx_enable(struct net_device *);
void rtl818x_tx_enable(struct net_device *);
void rtl818x_start_scanning(struct net_device *dev);
void rtl818x_start_scanning_s(struct net_device *dev);
void rtl818x_stop_scanning(struct net_device *dev);
void rtl818x_disassociate(struct net_device *dev);

//void fix_rx_fifo(struct net_device *dev);
void rtl818x_set_anaparam(struct net_device *dev, u32 a);
void rtl8185_set_anaparam2(struct net_device *dev, u32 a);
void rtl818x_set_hw_wep(struct net_device *dev);
void rtl818x_no_hw_wep(struct net_device *dev);
void rtl818x_set_mode(struct net_device *dev, int mode);
void rtl818x_update_msr(struct net_device *dev);

//void rtl818x_BSS_create(struct net_device *dev);
void rtl818x_beacon_tx_disable(struct net_device *dev);
void rtl818x_beacon_rx_disable(struct net_device *dev);
void rtl818x_conttx_enable(struct net_device *dev);
void rtl818x_conttx_disable(struct net_device *dev);
int rtl818x_down(struct net_device *dev);
int rtl818x_up(struct net_device *dev);
void rtl818x_commit(struct net_device *dev);
void rtl818x_set_chan(struct net_device *dev, short ch);
void rtl818x_set_master_essid(struct net_device *dev, char *essid);
void rtl818x_update_beacon_security(struct net_device *dev);
void write_phy(struct net_device *dev, u8 adr, u8 data);
void write_phy_cck(struct net_device *dev, u8 adr, u32 data);
void write_phy_ofdm(struct net_device *dev, u8 adr, u32 data);
void rtl8185_tx_antenna(struct net_device *dev, u8 ant);
void rtl8185_rf_pins_enable(struct net_device *dev);
void IBSS_randomize_cell(struct net_device *dev);
#endif	/* RTL818XH */
