/* $Id: rtl818x_wx.c,v 1.1 2006/06/30 18:52:54 bsd3 Exp $ */

/*- 
 * Driver for Realtek's RTL8180L/8185L 802.11(a)/b/g MAC/BBP chipset
 * based wireless cards, see README for more details.
 *
 * Copyright (C) Andrea Merello 2004  <andreamrl@tiscali.it> 
 *   
 * Parts of this driver are based on the GPL portion of the Realtek's 
 * official driver
 *   
 * Parts of this driver are based on the rtl8180 driver skeleton from
 * Patric Schenke, Andres Salomon and others.
 *   
 * Parts of this driver are based on the Intel Pro Wireless 2100 GPL
 * driver.
 *   
 * Thanks to the authors of all above said projects and the Ndiswrapper
 * project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

/* This file contains wireless extension handlers and is also part of
 * rtl8185 OpenSource driver */

#include "rtl818x.h"
#include "rtl818x_hw.h"
#include "rtl818x_sa2400.h"

#define RATE_COUNT 12
u32 rtl818x_rates[] = {
	1000000, 2000000, 5500000, 11000000,
	6000000, 9000000, 12000000, 18000000,
	24000000, 36000000, 48000000, 54000000
};

static int
rtl818x_wx_get_freq(struct net_device *dev,
		    struct iw_request_info *a, union iwreq_data *wrqu,
		    char *b)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	return ieee80211_wx_get_freq(priv->ieee80211, a, wrqu, b);
}

int
rtl818x_wx_set_key(struct net_device *dev, struct iw_request_info *info,
		   union iwreq_data *wrqu, char *key)
{
	struct iw_point *erq = &(wrqu->encoding);
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	// if (erq->flags & IW_ENCODE_DISABLED) {
	// i = erq->flags & IW_ENCODE_INDEX;
	// if (i < 1 || i > 4)
	// }

	if (erq->length > 0) {
		// int len = erq->length <= 5 ? 5 : 13;
		u32 *tkey = (u32 *) key;

		priv->key0[0] = tkey[0];
		priv->key0[1] = tkey[1];
		priv->key0[2] = tkey[2];
		priv->key0[3] = tkey[3] & 0xff;
		RTL818X_NOTICE("Setting wep key to 0x%x 0x%x 0x%x 0x%x",
			       tkey[0], tkey[1], tkey[2], tkey[3]);
		rtl818x_set_hw_wep(dev);
	}

	return 0;
}

static int
rtl818x_wx_set_beaconinterval(struct net_device *dev,
			      struct iw_request_info *aa,
			      union iwreq_data *wrqu, char *b)
{
	int *parms = (int *) b;
	int bi = parms[0];
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	RTL818X_NOTICE("setting beacon interval to 0x%x", bi);
	priv->ieee80211->current_network.beacon_interval = bi;
	rtl818x_commit(dev);

	up(&priv->wx_sem);
	return 0;
}

static int
rtl818x_wx_get_mode(struct net_device *dev, struct iw_request_info *a,
		    union iwreq_data *wrqu, char *b)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	return ieee80211_wx_get_mode(priv->ieee80211, a, wrqu, b);
}

static int
rtl818x_wx_get_rate(struct net_device *dev,
		    struct iw_request_info *info,
		    union iwreq_data *wrqu, char *extra)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	return ieee80211_wx_get_rate(priv->ieee80211, info, wrqu, extra);
}

static int
rtl818x_wx_set_rate(struct net_device *dev,
		    struct iw_request_info *info,
		    union iwreq_data *wrqu, char *extra)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	ret = ieee80211_wx_set_rate(priv->ieee80211, info, wrqu, extra);

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_set_crcmon(struct net_device *dev,
		      struct iw_request_info *info,
		      union iwreq_data *wrqu, char *extra)
{
	int *parms = (int *) extra;
	int enable = (parms[0] > 0);
	struct rtl818x_priv *priv = ieee80211_priv(dev);
	short prev = priv->crcmon;

	down(&priv->wx_sem);

	if (enable)
		priv->crcmon = 1;
	else
		priv->crcmon = 0;

	RTL818X_INFO("%sbad CRC in monitor mode are %s",
		     priv->crcmon ? "" : "(??) ",
		     priv->crcmon ? "accepted" : "rejected");

	if (prev != priv->crcmon && priv->up) {
		rtl818x_down(dev);
		rtl818x_up(dev);
	}

	up(&priv->wx_sem);
	return 0;
}

static int
rtl818x_wx_set_mode(struct net_device *dev, struct iw_request_info *a,
		    union iwreq_data *wrqu, char *b)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	ret = ieee80211_wx_set_mode(priv->ieee80211, a, wrqu, b);
	// rtl818x_commit(dev);

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_get_range(struct net_device *dev,
		     struct iw_request_info *info,
		     union iwreq_data *wrqu, char *extra)
{
	int i;
	struct iw_range *range = (struct iw_range *) extra;
	struct rtl818x_priv *priv = ieee80211_priv(dev);
	u16 val;

	wrqu->data.length = sizeof(*range);
	memset(range, 0, sizeof(*range));

	/* Let's try to keep this struct in the same order as in
	   linux/include/wireless.h */

	/* TODO: See what values we can set, and remove the ones we can't
	   set, or fill them with some default data. */

	/* ~5 Mb/s real (802.11b) */
	range->throughput = 5 * 1000 * 1000;

	// TODO: Not used in 802.11b?
	// range->min_nwid; /* Minimal NWID we are able to set */
	// TODO: Not used in 802.11b?
	// range->max_nwid; /* Maximal NWID we are able to set */

	/* Old Frequency (backward compat - moved lower ) */
	// range->old_num_channels; 
	// range->old_num_frequency;
	// range->old_freq[6]; /* Filler to keep "version" at the same offset 
	// */
	if (priv->rf_set_sens != NULL)
		range->sensitivity = priv->max_sens;	/* signal level
							   threshold range */
	range->max_qual.qual = 100;
	/* TODO: Find real max RSSI and stick here */
	range->max_qual.level = 0;
	range->max_qual.noise = -98;
	range->max_qual.updated = 7;	/* Updated all three */
	range->avg_qual.qual = 92;	/* > 8% missed beacons is 'bad' */
	/* TODO: Find real 'good' to 'bad' threshol value for RSSI */
	range->avg_qual.level = 20 + -98;
	range->avg_qual.noise = 0;
	range->avg_qual.updated = 7;	/* Updated all three */
	range->num_bitrates = RATE_COUNT;

	for (i = 0; i < RATE_COUNT && i < IW_MAX_BITRATES; i++) {
		range->bitrate[i] = rtl818x_rates[i];
	}

	range->min_frag = MIN_FRAG_THRESHOLD;
	range->max_frag = MAX_FRAG_THRESHOLD;
	range->pm_capa = 0;
	range->we_version_compiled = WIRELESS_EXT;
	range->we_version_source = 16;
	// range->retry_capa; /* What retry options are supported */
	// range->retry_flags; /* How to decode Max/Long Min/ShortRetryLimit
	// */
	// range->r_time_flags; /* How to decode Max/Long Min/ShortRetryLife
	// */
	// range->min_retry; /* Min/ShortRetryLimit */
	// range->max_retry; /* Max/LongRetryLimit */
	// range->min_r_time; /* Min/ShortRetryLife */
	// range->max_r_time; /* Max/LongRetryLife */
	range->num_channels = 14;

	for (i = 0, val = 0; i < 14; i++) {
		// Include only legal frequencies for some countries
		if ((priv->ieee80211->channel_map)[i + 1]) {
			range->freq[val].i = i + 1;
			range->freq[val].m =
			    ieee80211_wlan_frequencies[i] * 100000;
			range->freq[val].e = 1;
			val++;
		} else {
			// FIXME: do we need to set anything for channels
			// we don't use ?
		}

		if (val == IW_MAX_FREQUENCIES)
			break;
	}

	range->num_frequency = val;
	return 0;
}

static int
rtl818x_wx_set_scan(struct net_device *dev, struct iw_request_info *a,
		    union iwreq_data *wrqu, char *b)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	if (priv->up)
		ret = ieee80211_wx_set_scan(priv->ieee80211, a, wrqu, b);
	else
		ret = -1;

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_get_scan(struct net_device *dev, struct iw_request_info *a,
		    union iwreq_data *wrqu, char *b)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	if (priv->up)
		ret = ieee80211_wx_get_scan(priv->ieee80211, a, wrqu, b);
	else
		ret = -1;

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_set_essid(struct net_device *dev,
		     struct iw_request_info *a, union iwreq_data *wrqu,
		     char *b)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	ret = ieee80211_wx_set_essid(priv->ieee80211, a, wrqu, b);

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_get_essid(struct net_device *dev,
		     struct iw_request_info *a, union iwreq_data *wrqu,
		     char *b)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	ret = ieee80211_wx_get_essid(priv->ieee80211, a, wrqu, b);

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_set_freq(struct net_device *dev, struct iw_request_info *a,
		    union iwreq_data *wrqu, char *b)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	ret = ieee80211_wx_set_freq(priv->ieee80211, a, wrqu, b);

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_get_name(struct net_device *dev,
		    struct iw_request_info *info,
		    union iwreq_data *wrqu, char *extra)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	return ieee80211_wx_get_name(priv->ieee80211, info, wrqu, extra);
}

static int
rtl818x_wx_set_frag(struct net_device *dev,
		    struct iw_request_info *info,
		    union iwreq_data *wrqu, char *extra)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	if (wrqu->frag.disabled)
		priv->ieee80211->fts = MAX_FRAG_THRESHOLD;
	else {
		if (wrqu->frag.value < MIN_FRAG_THRESHOLD ||
		    wrqu->frag.value > MAX_FRAG_THRESHOLD)
			return -EINVAL;

		priv->ieee80211->fts = wrqu->frag.value & ~0x1;
	}

	return 0;
}

static int
rtl818x_wx_get_frag(struct net_device *dev,
		    struct iw_request_info *info,
		    union iwreq_data *wrqu, char *extra)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	wrqu->frag.value = priv->ieee80211->fts;
	wrqu->frag.fixed = 0;	/* no auto select */
	wrqu->frag.disabled = (wrqu->frag.value == MAX_FRAG_THRESHOLD);

	return 0;
}

static int
rtl818x_wx_set_wap(struct net_device *dev,
		   struct iw_request_info *info,
		   union iwreq_data *awrq, char *extra)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	ret = ieee80211_wx_set_wap(priv->ieee80211, info, awrq, extra);

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_get_wap(struct net_device *dev,
		   struct iw_request_info *info,
		   union iwreq_data *wrqu, char *extra)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	return ieee80211_wx_get_wap(priv->ieee80211, info, wrqu, extra);
}

static int
rtl818x_wx_set_enc(struct net_device *dev,
		   struct iw_request_info *info,
		   union iwreq_data *wrqu, char *key)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	if (priv->hw_wep)
		ret = rtl818x_wx_set_key(dev, info, wrqu, key);
	else {
		RTL818X_NOTICE("Setting SW wep key");
		ret =
		    ieee80211_wx_set_encode(priv->ieee80211, info, wrqu, key);
	}

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_get_enc(struct net_device *dev,
		   struct iw_request_info *info,
		   union iwreq_data *wrqu, char *key)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	return ieee80211_wx_get_encode(priv->ieee80211, info, wrqu, key);
}

static int
rtl818x_wx_set_scan_type(struct net_device *dev, struct iw_request_info *aa, union
			 iwreq_data *wrqu, char *p)
{
	int *parms = (int *) p;
	int mode = parms[0];
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	priv->ieee80211->active_scan = mode;

	return 1;
}

/* added by christian */
static int
rtl818x_wx_set_monitor_type(struct net_device *dev,
			    struct iw_request_info *aa, union
			    iwreq_data *wrqu, char *p)
{
	int *parms = (int *) p;
	int mode = parms[0];
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	if (priv->ieee80211->iw_mode != IW_MODE_MONITOR)
		return -1;

	priv->prism_hdr = mode;

	if (!mode)
		dev->type = ARPHRD_IEEE80211;
	else
		dev->type = ARPHRD_IEEE80211_PRISM;

	RTL818X_INFO("using %s RX encap", mode ? "AVS" : "80211");
	return 0;
}				/* end added christian */

static int
rtl818x_wx_set_retry(struct net_device *dev,
		     struct iw_request_info *info,
		     union iwreq_data *wrqu, char *extra)
{
	int err = 0;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	if (wrqu->retry.flags & IW_RETRY_LIFETIME || wrqu->retry.disabled) {
		err = -EINVAL;
		goto exit;
	}

	if (!(wrqu->retry.flags & IW_RETRY_LIMIT)) {
		err = -EINVAL;
		goto exit;
	}

	if (wrqu->retry.value > RTL818X_MAX_RETRY) {
		err = -EINVAL;
		goto exit;
	}

	if (wrqu->retry.flags & IW_RETRY_MAX) {
		priv->retry_rts = wrqu->retry.value;
		RTL818X_INFO("Setting retry for RTS/CTS data to %d",
			     wrqu->retry.value);

	} else {
		priv->retry_data = wrqu->retry.value;
		RTL818X_INFO("Setting retry for non RTS/CTS data to %d",
			     wrqu->retry.value);
	}

	/* FIXME ! We might try to write directly the TX config register or
	   to restart just the (R)TX process. I'm unsure if whole reset is
	   really needed */
	rtl818x_commit(dev);
	// if(priv->up) { 
	// rtl818x_rtx_disable(dev); 
	// rtl818x_rx_enable(dev);
	// rtl818x_tx_enable(dev);
	// }

      exit:
	up(&priv->wx_sem);
	return err;
}

static int
rtl818x_wx_get_retry(struct net_device *dev,
		     struct iw_request_info *info,
		     union iwreq_data *wrqu, char *extra)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	wrqu->retry.disabled = 0;	/* can't be disabled */

	if ((wrqu->retry.flags & IW_RETRY_TYPE) == IW_RETRY_LIFETIME)
		return -EINVAL;

	if (wrqu->retry.flags & IW_RETRY_MAX) {
		wrqu->retry.flags = IW_RETRY_LIMIT & IW_RETRY_MAX;
		wrqu->retry.value = priv->retry_rts;
	} else {
		wrqu->retry.flags = IW_RETRY_LIMIT & IW_RETRY_MIN;
		wrqu->retry.value = priv->retry_data;
	}

	// RTL818X_INFO("returning %d", wrqu->retry.value);
	return 0;
}

static int
rtl818x_wx_get_sens(struct net_device *dev,
		    struct iw_request_info *info,
		    union iwreq_data *wrqu, char *extra)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	if (priv->rf_set_sens == NULL)
		return -1;	/* we have not this support for this radio */

	wrqu->sens.value = priv->sens;
	return 0;
}

static int
rtl818x_wx_set_sens(struct net_device *dev,
		    struct iw_request_info *info,
		    union iwreq_data *wrqu, char *extra)
{
	short err = 0;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	// RTL818X_INFO("attempt to set sensivity to %ddb",
	// wrqu->sens.value);
	if (priv->rf_set_sens == NULL) {
		err = -1;	/* we have not this support for this radio */
		goto exit;
	}

	if (priv->rf_set_sens(dev, wrqu->sens.value) == 0)
		priv->sens = wrqu->sens.value;
	else
		err = -EINVAL;

      exit:
	up(&priv->wx_sem);
	return err;
}

static int
rtl818x_wx_set_rawtx(struct net_device *dev,
		     struct iw_request_info *info,
		     union iwreq_data *wrqu, char *extra)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	ret = ieee80211_wx_set_rawtx(priv->ieee80211, info, wrqu, extra);

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_get_power(struct net_device *dev,
		     struct iw_request_info *info,
		     union iwreq_data *wrqu, char *extra)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	ret = ieee80211_wx_get_power(priv->ieee80211, info, wrqu, extra);

	up(&priv->wx_sem);
	return ret;
}

static int
rtl818x_wx_set_power(struct net_device *dev,
		     struct iw_request_info *info,
		     union iwreq_data *wrqu, char *extra)
{
	int ret;
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	down(&priv->wx_sem);

	ret = ieee80211_wx_set_power(priv->ieee80211, info, wrqu, extra);

	up(&priv->wx_sem);
	return ret;
}

static int
dummy(struct net_device *dev, struct iw_request_info *a,
      union iwreq_data *wrqu, char *b)
{
	return -1;
}

static iw_handler rtl818x_wx_handlers[] = {
	NULL,			/* SIOCSIWCOMMIT */
	rtl818x_wx_get_name,	/* SIOCGIWNAME */
	dummy,			/* SIOCSIWNWID */
	dummy,			/* SIOCGIWNWID */
	rtl818x_wx_set_freq,	/* SIOCSIWFREQ */
	rtl818x_wx_get_freq,	/* SIOCGIWFREQ */
	rtl818x_wx_set_mode,	/* SIOCSIWMODE */
	rtl818x_wx_get_mode,	/* SIOCGIWMODE */
	rtl818x_wx_set_sens,	/* SIOCSIWSENS */
	rtl818x_wx_get_sens,	/* SIOCGIWSENS */
	NULL,			/* SIOCSIWRANGE */
	rtl818x_wx_get_range,	/* SIOCGIWRANGE */
	NULL,			/* SIOCSIWPRIV */
	NULL,			/* SIOCGIWPRIV */
	NULL,			/* SIOCSIWSTATS */
	NULL,			/* SIOCGIWSTATS */
	dummy,			/* SIOCSIWSPY */
	dummy,			/* SIOCGIWSPY */
	NULL,			/* SIOCGIWTHRSPY */
	NULL,			/* SIOCWIWTHRSPY */
	rtl818x_wx_set_wap,	/* SIOCSIWAP */
	rtl818x_wx_get_wap,	/* SIOCGIWAP */
	NULL,			/* -- hole -- */
	dummy,			/* SIOCGIWAPLIST -- depricated */
	rtl818x_wx_set_scan,	/* SIOCSIWSCAN */
	rtl818x_wx_get_scan,	/* SIOCGIWSCAN */
	rtl818x_wx_set_essid,	/* SIOCSIWESSID */
	rtl818x_wx_get_essid,	/* SIOCGIWESSID */
	dummy,			/* SIOCSIWNICKN */
	dummy,			/* SIOCGIWNICKN */
	NULL,			/* -- hole -- */
	NULL,			/* -- hole -- */
	rtl818x_wx_set_rate,	/* SIOCSIWRATE */
	rtl818x_wx_get_rate,	/* SIOCGIWRATE */
	dummy,			/* SIOCSIWRTS */
	dummy,			/* SIOCGIWRTS */
	rtl818x_wx_set_frag,	/* SIOCSIWFRAG */
	rtl818x_wx_get_frag,	/* SIOCGIWFRAG */
	dummy,			/* SIOCSIWTXPOW */
	dummy,			/* SIOCGIWTXPOW */
	rtl818x_wx_set_retry,	/* SIOCSIWRETRY */
	rtl818x_wx_get_retry,	/* SIOCGIWRETRY */
	rtl818x_wx_set_enc,	/* SIOCSIWENCODE */
	rtl818x_wx_get_enc,	/* SIOCGIWENCODE */
	rtl818x_wx_set_power,	/* SIOCSIWPOWER */
	rtl818x_wx_get_power,	/* SIOCGIWPOWER */
};

static const struct iw_priv_args rtl818x_private_args[] = {
	{
	 SIOCIWFIRSTPRIV + 0x0,
	 IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "badcrc"},
	{
	 SIOCIWFIRSTPRIV + 0x1,
	 IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "beaconint"},

	/* added by christian */
	// {
	// SIOCIWFIRSTPRIV + 0x2,
	// IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "prismhdr"
	// }, /* end added by christian */
	{
	 SIOCIWFIRSTPRIV + 0x3,
	 IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "activescan"},
	{
	 SIOCIWFIRSTPRIV + 0x4,
	 IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "rawtx"}
};

static iw_handler rtl818x_private_handler[] = {
	rtl818x_wx_set_crcmon,
	    /*SIOCIWSECONDPRIV*/ rtl818x_wx_set_beaconinterval,
	rtl818x_wx_set_monitor_type,
	rtl818x_wx_set_scan_type,
	rtl818x_wx_set_rawtx,
};

#if WIRELESS_EXT >= 17
static struct iw_statistics *
rtl818x_get_wireless_stats(struct net_device *dev)
{
	struct rtl818x_priv *priv = ieee80211_priv(dev);

	return &priv->wstats;
}
#endif

struct iw_handler_def rtl818x_wx_handlers_def = {
	.standard = rtl818x_wx_handlers,
	.num_standard = sizeof(rtl818x_wx_handlers) / sizeof(iw_handler),
	.private = rtl818x_private_handler,
	.num_private = sizeof(rtl818x_private_handler) / sizeof(iw_handler),
	.num_private_args =
	    sizeof(rtl818x_private_args) / sizeof(struct iw_priv_args),
#if WIRELESS_EXT >= 17
	.get_wireless_stats = rtl818x_get_wireless_stats,
#endif
	.private_args = (struct iw_priv_args *) rtl818x_private_args,
};
